"""This module contains the user interface for the prismatools package."""

import os
import math
import numpy as np
import xarray as xr
from typing import Optional

import ipyleaflet
import ipyevents
import ipywidgets as widgets
from ipyfilechooser import FileChooser
from ipyleaflet import WidgetControl, LayersControl

import bqplot
from bqplot import pyplot as plt

from .prisma import extract_prisma


def main_toolbar(m):
    """Creates the main toolbar and adds it to the map.

    Args:
        m (leafmap.Map): The leafmap Map object.
    """
    # add here additional tools
    all_tools = {
        "map": {
            "name": "basemap",
            "tooltip": "Change basemap",
        },
    }

    tools = {}
    for tool in all_tools:
        if os.environ.get(all_tools[tool]["name"].upper(), "").upper() != "FALSE":
            tools[tool] = all_tools[tool]

    icons = list(tools.keys())
    tooltips = [item["tooltip"] for item in list(tools.values())]
    toolnames = [item["name"].upper() for item in list(all_tools.values())]
    toolnames.sort()
    setattr(m, "_ENV_VARS", toolnames)

    icon_width = "32px"
    icon_height = "32px"
    n_cols = 3
    n_rows = math.ceil(len(icons) / n_cols)

    toolbar_grid = widgets.GridBox(
        children=[
            widgets.ToggleButton(
                layout=widgets.Layout(
                    width="auto", height="auto", padding="0px 0px 0px 4px"
                ),
                button_style="primary",
                icon=icons[i],
                tooltip=tooltips[i],
            )
            for i in range(len(icons))
        ],
        layout=widgets.Layout(
            width="109px",
            grid_template_columns=(icon_width + " ") * n_cols,
            grid_template_rows=(icon_height + " ") * n_rows,
            grid_gap="1px 1px",
            padding="5px",
        ),
    )
    m.toolbar = toolbar_grid

    def tool_callback(change):
        if change["new"]:
            current_tool = change["owner"]
            for tool in toolbar_grid.children:
                if tool is not current_tool:
                    tool.value = False
            tool = change["owner"]
            tool_name = tools[tool.icon]["name"]

            if tool_name == "basemap":
                change_basemap(m)
        else:
            # tool = change["owner"]
            # tool_name = tools[tool.icon]["name"]
            pass

        m.toolbar_reset()

    for tool in toolbar_grid.children:
        tool.observe(tool_callback, "value")

    toolbar_button = widgets.ToggleButton(
        value=False,
        tooltip="Toolbar",
        icon="wrench",
        layout=widgets.Layout(width="28px", height="28px", padding="0px 0px 0px 4px"),
    )
    m.toolbar_button = toolbar_button

    layers_button = widgets.ToggleButton(
        value=False,
        tooltip="Layers",
        icon="server",
        layout=widgets.Layout(height="28px", width="72px"),
    )

    toolbar_widget = widgets.VBox()
    toolbar_widget.children = [toolbar_button]
    toolbar_header = widgets.HBox()
    toolbar_header.children = [layers_button, toolbar_button]
    toolbar_footer = widgets.VBox()
    toolbar_footer.children = [toolbar_grid]

    toolbar_event = ipyevents.Event(
        source=toolbar_widget, watched_events=["mouseenter", "mouseleave"]
    )

    def handle_toolbar_event(event):
        if event["type"] == "mouseenter":
            toolbar_widget.children = [toolbar_header, toolbar_footer]
        elif event["type"] == "mouseleave":
            if not toolbar_button.value:
                toolbar_widget.children = [toolbar_button]
                toolbar_button.value = False
                layers_button.value = False

    toolbar_event.on_dom_event(handle_toolbar_event)

    def toolbar_btn_click(change):
        if change["new"]:
            layers_button.value = False
            toolbar_widget.children = [toolbar_header, toolbar_footer]
        else:
            if not layers_button.value:
                toolbar_widget.children = [toolbar_button]

    toolbar_button.observe(toolbar_btn_click, "value")

    def layers_btn_click(change):
        if change["new"]:
            toolbar_footer.children = layer_manager_gui(m, return_widget=True)
            toolbar_button.value = False
        else:
            toolbar_footer.children = [toolbar_grid]

    layers_button.observe(layers_btn_click, "value")

    toolbar_control = WidgetControl(widget=toolbar_widget, position="topright")

    m.add(toolbar_control)


def change_basemap(m, position: Optional[str] = "topright"):
    """Widget for changing basemaps.

    Args:
        m (object): leafmap.Map.
    """
    from leafmap.basemaps import get_xyz_dict
    from leafmap import basemaps, get_basemap

    xyz_dict = get_xyz_dict()

    value = "OpenStreetMap"

    dropdown = widgets.Dropdown(
        options=list(basemaps.keys()),
        value=value,
        layout=widgets.Layout(width="200px"),
    )

    close_btn = widgets.Button(
        icon="times",
        tooltip="Close the basemap widget",
        button_style="primary",
        layout=widgets.Layout(width="32px"),
    )

    basemap_widget = widgets.HBox([dropdown, close_btn])

    def on_click(change):
        if change["new"]:
            basemap_name = dropdown.value
            if basemap_name not in m.get_layer_names():
                m.add_basemap(basemap_name)
                if basemap_name in xyz_dict:
                    if "bounds" in xyz_dict[basemap_name]:
                        bounds = xyz_dict[basemap_name]["bounds"]
                        bounds = [
                            bounds[0][1],
                            bounds[0][0],
                            bounds[1][1],
                            bounds[1][0],
                        ]
                        m.zoom_to_bounds(bounds)

    dropdown.observe(on_click, "value")

    def close_click(change):
        m.toolbar_reset()
        if m.basemap_ctrl is not None and m.basemap_ctrl in m.controls:
            m.remove_control(m.basemap_ctrl)
        basemap_widget.close()

    close_btn.on_click(close_click)

    basemap_control = WidgetControl(widget=basemap_widget, position=position)
    m.add(basemap_control)
    m.basemap_ctrl = basemap_control


def layer_manager_gui(
    m,
    position: Optional[str] = "topright",
    opened: Optional[bool] = True,
    return_widget: Optional[bool] = False,
):
    """Creates a layer manager widget.

    Args:
        m (geemap.Map): The geemap.Map object.
        position (str, optional): The position of the widget. Defaults to "topright".
        return_widget (bool, optional): Whether to return the widget. Defaults to False.
    """

    layers_button = widgets.ToggleButton(
        value=False,
        tooltip="Layer Manager",
        icon="server",
        layout=widgets.Layout(width="28px", height="28px", padding="0px 0px 0px 4px"),
    )

    close_button = widgets.ToggleButton(
        value=False,
        tooltip="Close the tool",
        icon="times",
        button_style="primary",
        layout=widgets.Layout(height="28px", width="28px", padding="0px 0px 0px 4px"),
    )

    toolbar_header = widgets.HBox()
    toolbar_header.children = [layers_button]
    toolbar_footer = widgets.VBox()
    toolbar_footer.children = []
    toolbar_widget = widgets.VBox()
    toolbar_widget.children = [toolbar_header]

    def toolbar_btn_click(change):
        if change["new"]:
            close_button.value = False
            toolbar_widget.children = [toolbar_header, toolbar_footer]
        else:
            if not close_button.value:
                toolbar_widget.children = [layers_button]

    layers_button.observe(toolbar_btn_click, "value")

    def close_btn_click(change):
        if change["new"]:
            layers_button.value = False
            m.toolbar_reset()
            if m.layer_manager is not None and m.layer_manager in m.controls:
                m.remove_control(m.layer_manager)
                m.layer_manager = None
            toolbar_widget.close()

    close_button.observe(close_btn_click, "value")

    def layers_btn_click(change):
        if change["new"]:
            layers_hbox = []
            all_layers_chk = widgets.Checkbox(
                value=False,
                description="All layers on/off",
                indent=False,
                layout=widgets.Layout(height="18px", padding="0px 8px 25px 8px"),
            )
            all_layers_chk.layout.width = "30ex"
            layers_hbox.append(all_layers_chk)

            def all_layers_chk_changed(change):
                if change["new"]:
                    for layer in m.layers:
                        if hasattr(layer, "visible"):
                            layer.visible = True
                else:
                    for layer in m.layers:
                        if hasattr(layer, "visible"):
                            layer.visible = False

            all_layers_chk.observe(all_layers_chk_changed, "value")

            layers = [lyr for lyr in m.layers if lyr.name]

            # if the layers contain unsupported layers (e.g., GeoJSON, GeoData), adds the ipyleaflet built-in LayerControl
            if len(layers) < (len(m.layers) - 1):
                if m.layer_control is None:
                    layer_control = LayersControl(position="topright")
                    m.layer_control = layer_control
                if m.layer_control not in m.controls:
                    m.add(m.layer_control)

            # for non-TileLayer, use layer.style={'opacity':0, 'fillOpacity': 0} to turn layer off.
            for layer in layers:
                visible = True
                if hasattr(layer, "visible"):
                    visible = layer.visible
                layer_chk = widgets.Checkbox(
                    value=visible,
                    description=layer.name,
                    indent=False,
                    layout=widgets.Layout(height="18px"),
                )
                layer_chk.layout.width = "25ex"

                if layer in m.geojson_layers:
                    try:
                        opacity = max(
                            layer.style["opacity"], layer.style["fillOpacity"]
                        )
                    except KeyError:
                        opacity = 1.0
                elif hasattr(layer, "opacity"):
                    opacity = layer.opacity
                else:
                    opacity = 1.0

                layer_opacity = widgets.FloatSlider(
                    value=opacity,
                    min=0,
                    max=1,
                    step=0.01,
                    readout=False,
                    layout=widgets.Layout(width="80px"),
                )
                layer_settings = widgets.Button(
                    icon="gear",
                    tooltip=layer.name,
                    layout=widgets.Layout(
                        width="25px", height="25px", padding="0px 0px 0px 0px"
                    ),
                )

                def layer_settings_click(b):
                    if b.tooltip in m.cog_layer_dict:
                        m._add_layer_editor(
                            position="topright",
                            layer_dict=m.cog_layer_dict[b.tooltip],
                        )

                layer_settings.on_click(layer_settings_click)

                def layer_opacity_changed(change):
                    if change["new"]:
                        layer.style = {
                            "opacity": change["new"],
                            "fillOpacity": change["new"],
                        }

                if hasattr(layer, "visible"):
                    widgets.jslink((layer_chk, "value"), (layer, "visible"))

                if layer in m.geojson_layers:
                    layer_opacity.observe(layer_opacity_changed, "value")
                elif hasattr(layer, "opacity"):
                    widgets.jsdlink((layer_opacity, "value"), (layer, "opacity"))

                hbox = widgets.HBox(
                    [layer_chk, layer_settings, layer_opacity],
                    layout=widgets.Layout(padding="0px 8px 0px 8px"),
                )
                layers_hbox.append(hbox)
                m.layer_widget = layers_hbox

            toolbar_header.children = [close_button, layers_button]
            toolbar_footer.children = layers_hbox

        else:
            toolbar_header.children = [layers_button]

    layers_button.observe(layers_btn_click, "value")
    layers_button.value = opened

    if not hasattr(m, "_layer_manager_widget"):
        m._layer_manager_widget = toolbar_footer

    if return_widget:
        return m.layer_widget
    else:
        layer_control = WidgetControl(widget=toolbar_widget, position=position)

        if layer_control not in m.controls:
            m.add_control(layer_control)
            m.layer_manager = layer_control


class spectralWidget(widgets.HBox):
    """
    A widget for spectral data visualization on a map.

    Attributes:
        _host_map (Map): The map to host the widget.
        on_close (function): Function to be called when the widget is closed.
        _output_widget (widgets.Output): The output widget to display results.
        _output_control (ipyleaflet.WidgetControl): The control for the output widget.
        _on_map_interaction (function): Function to handle map interactions.
        _spectral_widget (SpectralWidget): The spectral widget itself.
        _spectral_control (ipyleaflet.WidgetControl): The control for the spectral widget.
    """

    def __init__(
        self, host_map, stack=True, position="topright", xlim=None, ylim=None, **kwargs
    ):
        """
        Initializes a new instance of the SpectralWidget class.

        Args:
            host_map (Map): The map to host the widget.
            stack (bool, optional): Whether to stack the plots. Defaults to True.
            position (str, optional): The position of the widget on the map. Defaults to "topright".
            xlim (tuple, optional): The x-axis limits. Defaults to None.
            ylim (tuple, optional): The y-axis limits. Defaults to None.
        """
        self._host_map = host_map

        if not hasattr(self._host_map, "_spectral_counter"):
            self._host_map._spectral_counter = 0

        self.on_close = None
        self._stack = stack
        self._show_plot = False

        fig_margin = {"top": 20, "bottom": 35, "left": 50, "right": 20}
        fig = plt.figure(
            # title=None,
            fig_margin=fig_margin,
            layout={"width": "500px", "height": "300px"},
        )

        self._fig = fig
        self._host_map._fig = fig

        layer_names = list(host_map.cog_layer_dict.keys())
        layers_widget = widgets.Dropdown(options=layer_names)
        layers_widget.layout.width = "18ex"

        close_btn = widgets.Button(
            icon="times",
            tooltip="Close the widget",
            button_style="primary",
            layout=widgets.Layout(width="32px"),
        )

        reset_btn = widgets.Button(
            icon="trash",
            tooltip="Remove all markers",
            button_style="primary",
            layout=widgets.Layout(width="32px"),
        )

        stack_btn = widgets.ToggleButton(
            value=stack,
            icon="area-chart",
            tooltip="Stack spectral signatures",
            button_style="primary",
            layout=widgets.Layout(width="32px"),
        )

        def reset_btn_click(_):
            if hasattr(self._host_map, "_plot_marker_cluster"):
                self._host_map._plot_marker_cluster.markers = []
                self._host_map._plot_markers = []

            if hasattr(self._host_map, "_spectral_data"):
                self._host_map._spectral_data = {}

            if hasattr(self._host_map, "_spectral_counter"):
                self._host_map._spectral_counter = 0

            self._output_widget.clear_output()
            self._show_plot = False
            plt.clear()

        reset_btn.on_click(reset_btn_click)

        save_btn = widgets.Button(
            icon="floppy-o",
            tooltip="Save the data to a CSV",
            button_style="primary",
            layout=widgets.Layout(width="32px"),
        )

        def chooser_callback(chooser):
            if chooser.selected:
                file_path = chooser.selected
                self._host_map.spectral_to_csv(file_path)
                if (
                    hasattr(self._host_map, "_file_chooser_control")
                    and self._host_map._file_chooser_control in self._host_map.controls
                ):
                    self._host_map.remove_control(self._host_map._file_chooser_control)
                    self._host_map._file_chooser.close()

        def save_btn_click(_):
            if not hasattr(self._host_map, "_spectral_data"):
                return

            self._output_widget.clear_output()
            file_chooser = FileChooser(
                os.getcwd(), layout=widgets.Layout(width="454px")
            )
            file_chooser.filter_pattern = "*.csv"
            file_chooser.use_dir_icons = True
            file_chooser.title = "Save spectral data to a CSV file"
            file_chooser.default_filename = "spectral_data.csv"
            file_chooser.show_hidden = False
            file_chooser.register_callback(chooser_callback)
            file_chooser_control = ipyleaflet.WidgetControl(
                widget=file_chooser, position="topright"
            )
            self._host_map.add(file_chooser_control)
            setattr(self._host_map, "_file_chooser", file_chooser)
            setattr(self._host_map, "_file_chooser_control", file_chooser_control)

        save_btn.on_click(save_btn_click)

        def close_widget(_):
            self.cleanup()

        close_btn.on_click(close_widget)

        super().__init__([layers_widget, stack_btn, reset_btn, save_btn, close_btn])

        output = widgets.Output()
        output_control = ipyleaflet.WidgetControl(widget=output, position="bottomright")
        self._output_widget = output
        self._output_control = output_control
        self._host_map.add(output_control)

        if not hasattr(self._host_map, "_spectral_data"):
            self._host_map._spectral_data = {}

        def handle_interaction(**kwargs):

            latlon = kwargs.get("coordinates")
            lat = latlon[0]
            lon = latlon[1]
            if kwargs.get("type") == "click" and self._host_map._layer_editor is None:

                self._host_map._spectral_counter += 1
                idx = self._host_map._spectral_counter
                label_txt = f"P {idx}"

                layer_name = layers_widget.value

                if not hasattr(self._host_map, "_plot_markers"):
                    self._host_map._plot_markers = []

                markers = self._host_map._plot_markers
                marker_cluster = self._host_map._plot_marker_cluster
                marker = ipyleaflet.Marker(location=latlon, draggable=False)
                marker.popup = widgets.HTML(value=label_txt)
                markers.append(marker)
                marker_cluster.markers = markers
                self._host_map._plot_marker_cluster = marker_cluster

                xlabel = "Wavelength (nm)"
                ylabel = "Reflectance"

                ds = self._host_map.cog_layer_dict[layer_name]["xds"]

                if self._host_map.cog_layer_dict[layer_name]["hyper"] == "PRISMA":
                    da = extract_prisma(ds, lat, lon)

                self._host_map._spectral_data[f"({lat:.4f} {lon:.4f})"] = da.values

                if self._host_map.cog_layer_dict[layer_name]["hyper"] != "XARRAY":
                    da[da < 0] = np.nan
                    x_axis_options = {"label_offset": "30px"}
                else:
                    x_axis_options = {
                        "label_offset": "30px",
                        "tick_format": "0d",
                        "num_ticks": da.sizes["band"],
                    }
                axes_options = {
                    "x": x_axis_options,
                    "y": {"label_offset": "35px"},
                }

                if not stack_btn.value:
                    plt.clear()
                    plt.plot(
                        da.coords[da.dims[0]].values,
                        da.values,
                        axes_options=axes_options,
                    )
                else:
                    color = np.random.rand(
                        3,
                    )
                    if "wavelength" in da.coords:
                        xlabel = "Wavelength (nm)"
                        x_values = da["wavelength"].values
                    else:
                        xlabel = "Band"
                        x_values = da.coords[da.dims[0]].values

                    line = plt.plot(
                        x_values,
                        da.values,
                        color=color,
                        axes_options=axes_options,
                    )
                    line.labels = [label_txt]

                    plt.legend()

                    try:
                        if isinstance(self._fig.axes[0], bqplot.ColorAxis):
                            self._fig.axes = self._fig.axes[1:]
                        elif isinstance(self._fig.axes[-1], bqplot.ColorAxis):
                            self._fig.axes = self._fig.axes[:-1]
                    except Exception:
                        pass

                plt.xlabel(xlabel)
                plt.ylabel(ylabel)
                if xlim:
                    plt.xlim(xlim[0], xlim[1])
                if ylim:
                    plt.ylim(ylim[0], ylim[1])

                if not self._show_plot:
                    with self._output_widget:
                        plt.show()
                        self._show_plot = True

                self._host_map.default_style = {"cursor": "crosshair"}

        self._host_map.on_interaction(handle_interaction)
        self._on_map_interaction = handle_interaction

        self._spectral_widget = self
        self._spectral_control = ipyleaflet.WidgetControl(
            widget=self, position=position
        )
        self._host_map.add(self._spectral_control)

    def cleanup(self):
        """Removes the widget from the map and performs cleanup."""
        if self._host_map:
            self._host_map.default_style = {"cursor": "default"}
            self._host_map.on_interaction(self._on_map_interaction, remove=True)

            if self._output_control:
                self._host_map.remove_control(self._output_control)

                if self._output_widget:
                    self._output_widget.close()
                    self._output_widget = None

            if self._spectral_control:
                self._host_map.remove_control(self._spectral_control)
                self._spectral_control = None

                if self._spectral_widget:
                    self._spectral_widget.close()
                    self._spectral_widget = None

            if hasattr(self._host_map, "_plot_marker_cluster"):
                self._host_map._plot_marker_cluster.markers = []
                self._host_map._plot_markers = []

            if hasattr(self._host_map, "_spectral_data"):
                self._host_map._spectral_data = {}

            if hasattr(self._host_map, "_spectral_counter"):
                self._host_map._spectral_counter = 0

            if hasattr(self, "_output_widget") and self._output_widget is not None:
                self._output_widget.clear_output()

        if self.on_close is not None:
            self.on_close()



"""The prisma module provides functions to read and write PRISMA L2 data products"""

# TODO: check if "extract_prisma" works also for panchromatic data.

import os
import numpy as np
import rasterio
import xarray as xr
import rioxarray
import h5py
from typing import List, Tuple, Union, Optional, Any

from .utils import check_valid_file, convert_coords, get_transform


def read_prismaL2D(
    file_path: str,
    wavelengths: Optional[List[float]] = None,
    method: str = "nearest",
    panchromatic: bool = False,
    extract_error: bool = False,
) -> xr.Dataset:
    """
    The function reads PRISMA Level-2D .he5 data (hyperspectral or panchromatic),
    applies digital number to reflectance scaling, handles fill values, optionally
    selects specific wavelengths, and builds an xarray.Dataset with spatial
    coordinates and CRS information. Optionally reads pixel-level error matrices
    when available.

    Args:
        file_path (str):
            Path to the PRISMA L2D .he5 file.
        wavelengths (Optional[List[float]], optional):
            List of wavelengths (in nm) to extract (hyperspectral only).
            - If None, all available wavelengths are used.
            - If provided, can select by exact match or nearest available wavelength
              depending on `method`.
        method (str, optional):
            Wavelength selection method (used if `wavelengths` is provided, hyperspectral only):
            - "nearest" (default): select the closest available band for each requested wavelength.
            - "exact": select only exact matches; raises ValueError if none found.
        panchromatic (bool, optional):
            If True, read the panchromatic cube instead of the hyperspectral VNIR+SWIR cubes.
            Defaults to False.
        extract_error (bool, optional):
            If True, attempt to read and attach pixel-level error matrix
            fields from the product (when present). Defaults to False.

    Returns:
        xr.Dataset:
        An xarray.Dataset containing reflectance data with dimensions:
            - Hyperspectral: ("y", "x", "wavelength")
            - Panchromatic: ("y", "x")

        The dataset may include an optional variable named `error_matrix` when
        `extract_error=True` and pixel-level error matrices are present in the
        product. The shape of `error_matrix` matches the reflectance dims.
    """
    # check if file is valid
    if not check_valid_file(file_path, type="PRS_L2D"):
        raise ValueError(f"{file_path} is not a valid PRS_L2D file or does not exist.")

    try:
        with h5py.File(file_path, "r") as f:
            epsg_code = f.attrs["Epsg_Code"][()]
            ul_easting = f.attrs["Product_ULcorner_easting"][()]
            ul_northing = f.attrs["Product_ULcorner_northing"][()]

            if panchromatic:
                # --- PANCHROMATIC ---
                pancube_path = "HDFEOS/SWATHS/PRS_L2D_PCO/Data Fields/Cube"
                pancube_data = f[pancube_path][()]
                l2_scale_pan_min = f.attrs["L2ScalePanMin"][()]
                l2_scale_pan_max = f.attrs["L2ScalePanMax"][()]
                fill_value = 0
                max_data_value = 65535

                pancube_data = l2_scale_pan_min + (
                    pancube_data.astype(np.float32) / max_data_value
                ) * (l2_scale_pan_max - l2_scale_pan_min)
                pancube_data[pancube_data == fill_value] = np.nan

                rows, cols = pancube_data.shape
                transform = get_transform(ul_easting, ul_northing, res=5)
                x_coords = np.array([transform * (i, 0) for i in range(cols)])[:, 0]
                y_coords = np.array([transform * (0, j) for j in range(rows)])[:, 1]

                ds = xr.Dataset(
                    data_vars=dict(
                        reflectance=(
                            ["y", "x"],
                            pancube_data,
                            dict(
                                units="unitless",
                                _FillValue=np.nan,
                                standard_name="reflectance",
                                long_name="Panchromatic reflectance",
                            ),
                        ),
                    ),
                    coords=dict(
                        y=(["y"], y_coords, dict(units="m")),
                        x=(["x"], x_coords, dict(units="m")),
                    ),
                )
                if extract_error:
                    try:
                        err_path = (
                            "HDFEOS/SWATHS/PRS_L2D_PCO/Data Fields/PIXEL_L2_ERR_MATRIX"
                        )
                        err_arr = f[err_path][()].astype(np.float32)
                        # mask error where DN fill is present
                        try:
                            err_arr[np.isnan(pancube_data)] = np.nan
                        except Exception:
                            pass
                        ds["error_matrix"] = ("y", "x"), err_arr
                    except Exception:
                        # no error matrix available; continue silently
                        pass

            else:
                # --- HYPERSPECTRAL CUBE ---
                swir_cube = f["HDFEOS/SWATHS/PRS_L2D_HCO/Data Fields/SWIR_Cube"][()]
                vnir_cube = f["HDFEOS/SWATHS/PRS_L2D_HCO/Data Fields/VNIR_Cube"][()]
                vnir_wavelengths = f.attrs["List_Cw_Vnir"][()]
                swir_wavelengths = f.attrs["List_Cw_Swir"][()]
                l2_scale_vnir_min = f.attrs["L2ScaleVnirMin"][()]
                l2_scale_vnir_max = f.attrs["L2ScaleVnirMax"][()]
                l2_scale_swir_min = f.attrs["L2ScaleSwirMin"][()]
                l2_scale_swir_max = f.attrs["L2ScaleSwirMax"][()]
                fill_value = 0
                max_data_value = 65535

                vnir_cube = l2_scale_vnir_min + (
                    vnir_cube.astype(np.float32) / max_data_value
                ) * (l2_scale_vnir_max - l2_scale_vnir_min)
                swir_cube = l2_scale_swir_min + (
                    swir_cube.astype(np.float32) / max_data_value
                ) * (l2_scale_swir_max - l2_scale_swir_min)

                vnir_cube[vnir_cube == fill_value] = np.nan
                swir_cube[swir_cube == fill_value] = np.nan

                full_cube = np.concatenate((vnir_cube, swir_cube), axis=1)
                full_wavelengths = np.concatenate((vnir_wavelengths, swir_wavelengths))

                # filter and sort wavelengths
                valid_idx = full_wavelengths > 0
                full_wavelengths = full_wavelengths[valid_idx]
                full_cube = full_cube[:, valid_idx, :]
                sort_idx = np.argsort(full_wavelengths)
                full_wavelengths = full_wavelengths[sort_idx]
                full_cube = full_cube[:, sort_idx, :]

                # select requested wavelengths
                if wavelengths is not None:
                    requested = np.array(wavelengths)
                    if method == "exact":
                        idx = np.where(np.isin(full_wavelengths, requested))[0]
                        if len(idx) == 0:
                            raise ValueError(
                                "No requested wavelengths found (exact match)."
                            )
                    else:
                        idx = np.array(
                            [np.abs(full_wavelengths - w).argmin() for w in requested]
                        )
                    full_wavelengths = full_wavelengths[idx]
                    full_cube = full_cube[:, idx, :]

                rows, cols = full_cube.shape[0], full_cube.shape[2]
                transform = get_transform(ul_easting, ul_northing, res=30)
                x_coords = np.array([transform * (i, 0) for i in range(cols)])[:, 0]
                y_coords = np.array([transform * (0, j) for j in range(rows)])[:, 1]

                ds = xr.Dataset(
                    data_vars=dict(
                        reflectance=(
                            ["y", "wavelength", "x"],
                            full_cube,
                            dict(
                                units="unitless",
                                _FillValue=np.nan,
                                standard_name="reflectance",
                                long_name="Combined atmospherically corrected surface reflectance",
                            ),
                        ),
                    ),
                    coords=dict(
                        wavelength=(
                            ["wavelength"],
                            full_wavelengths,
                            dict(long_name="center wavelength", units="nm"),
                        ),
                        y=(["y"], y_coords, dict(units="m")),
                        x=(["x"], x_coords, dict(units="m")),
                    ),
                )
                if extract_error:
                    try:
                        v_err_path = "HDFEOS/SWATHS/PRS_L2D_HCO/Data Fields/VNIR_PIXEL_L2_ERR_MATRIX"
                        s_err_path = "HDFEOS/SWATHS/PRS_L2D_HCO/Data Fields/SWIR_PIXEL_L2_ERR_MATRIX"
                        v_err = f[v_err_path][()].astype(np.float32)
                        s_err = f[s_err_path][()].astype(np.float32)
                        # mask raw DN fill values if raw arrays exist
                        try:
                            v_err[np.isnan(vnir_cube)] = np.nan
                        except Exception:
                            pass
                        try:
                            s_err[np.isnan(swir_cube)] = np.nan
                        except Exception:
                            pass

                        err_full = np.concatenate((v_err, s_err), axis=1)
                        # apply same wavelength filtering and sorting as reflectance
                        err_full = err_full[:, valid_idx, :]
                        err_full = err_full[:, sort_idx, :]
                        if wavelengths is not None:
                            err_full = err_full[:, idx, :]
                        ds["error_matrix"] = ("y", "wavelength", "x"), err_full
                    except Exception:
                        pass
                ds["reflectance"] = ds.reflectance.transpose("y", "x", "wavelength")
                # transpose error matrix as well to match reflectance dims
                try:
                    if "error_matrix" in ds:
                        ds["error_matrix"] = ds.error_matrix.transpose(
                            "y", "x", "wavelength"
                        )
                except Exception:
                    pass

    except Exception as e:
        raise RuntimeError(f"Error reading the file {file_path}: {e}")

    # write CRS and transform
    crs = f"EPSG:{epsg_code}"
    ds.rio.write_crs(crs, inplace=True)
    ds.rio.write_transform(transform, inplace=True)

    # global attributes
    ds.attrs.update(
        dict(
            units="unitless",
            _FillValue=0,
            grid_mapping="crs",
            standard_name="reflectance",
            Conventions="CF-1.6",
            crs=ds.rio.crs.to_string(),
        )
    )

    return ds


def write_prismaL2D(
    dataset: Union[xr.Dataset, str],
    output: str,
    panchromatic: bool = False,
    wavelengths: Optional[np.ndarray] = None,
    method: str = "nearest",
    **kwargs: Any,
) -> Optional[str]:
    """
    Write a PRISMA Level-2D dataset to a georeferenced raster file.

    This function takes a PRISMA Level-2D hyperspectral or panchromatic dataset
    (or the path to a `.he5` file), extracts reflectance data and georeferencing
    information, and saves it as a raster image (GeoTIFF by default).

    Args:
        dataset (Union[xr.Dataset, str]):
            PRISMA dataset as an `xarray.Dataset` or the path to a Level-2D `.he5` file.
        output (str):
            Path to the output raster file.
        panchromatic (bool, optional):
            If True, treat the dataset as a single-band panchromatic image.
            Defaults to False.
        wavelengths (Optional[np.ndarray], optional):
            Wavelengths (in nm) to select from the dataset (hyperspectral only).
            If None, all available wavelengths are included.
            Defaults to None.
        method (str, optional):
            Wavelength selection method (hyperspectral only):
                - `"nearest"`: select closest available bands.
                - `"exact"`: select exact matches only.
            Defaults to `"nearest"`.
        **kwargs (Any):
            Additional keyword arguments passed to:
                - `array_to_image()` for raster writing.
                - `rasterio.open()` for file creation.

    Returns:
        str: Output file path, or None if all values are NaN.
    """
    # load dataset if it's a path to .he5
    if isinstance(dataset, str):
        dataset = read_prismaL2D(dataset, panchromatic=panchromatic)

    # get np.array
    array = dataset["reflectance"].values
    if not np.any(np.isfinite(array)):
        print("Warning: All reflectance values are NaN. Output image will be blank.")
        return None

    # get band names (wavelength) and, eventually, select specific bands
    if array.ndim == 2:  # panchromatic
        kwargs["band_description"] = "Panchromatic band"
    else:  # cube
        if wavelengths is not None:
            dataset = dataset.sel(wavelength=wavelengths, method=method)
            array = dataset["reflectance"].values
        kwargs["wavelengths"] = dataset["wavelength"].values

    return array_to_image(
        array,
        output=output,
        transpose=False,
        crs=dataset.rio.crs,
        transform=dataset.rio.transform(),
        **kwargs,
    )


def extract_prisma(
    dataset: xr.Dataset,
    lat: float,
    lon: float,
    offset: float = 15.0,
) -> xr.DataArray:
    """
    Extracts an averaged reflectance spectrum from a PRISMA hyperspectral dataset.

    A square spatial window is centered at the given latitude and longitude.
    The reflectance values within that window are averaged across the spatial
    dimensions, producing a single spectrum.

    Args:
        dataset (xr.Dataset):
            PRISMA dataset containing a variable named "reflectance", with
            valid CRS information and shape (x, y, wavelength).
        lat (float):
            Latitude of the center point (in WGS84).
        lon (float):
            Longitude of the center point (in WGS84).
        offset (float, optional):
            Half-size of the square window for extraction, expressed in the
            dataset's projected coordinate units (e.g., meters).
            Defaults to `15.0`.

    Returns:
        xarray.DataArray: A 1D array containing the averaged reflectance values
        across wavelengths. If no matching pixels are found, returns NaN values.
    """
    if dataset.rio.crs is None:
        raise ValueError("Dataset CRS not set. Please provide dataset with CRS info.")

    crs = dataset.rio.crs.to_string()
    # convert lat/lon to projected coords
    x_proj, y_proj = convert_coords([(lat, lon)], "epsg:4326", crs)[0]

    da = dataset["reflectance"]
    x_con = (da["x"] > x_proj - offset) & (da["x"] < x_proj + offset)
    y_con = (da["y"] > y_proj - offset) & (da["y"] < y_proj + offset)

    data = da.where(x_con & y_con, drop=True)

    if "wavelength" in da.dims:
        data = data.mean(dim=["x", "y"], skipna=True)
        return xr.DataArray(
            data,
            dims=["wavelength"],
            coords={"wavelength": dataset.coords["wavelength"]},
        )
    else:
        # panchromatic
        data = data.mean(dim=["x", "y"], skipna=True)
        return xr.DataArray(
            [data.item()] if np.ndim(data) == 0 else data,
            dims=["value"],
        )


def array_to_image(
    array: np.ndarray,
    output: str,
    dtype: Optional[np.dtype] = None,
    compress: str = "lzw",
    transpose: bool = True,
    crs: Optional[str] = None,
    transform: Optional[tuple] = None,
    driver: str = "GTiff",
    **kwargs,
) -> str:
    """
    Save a NumPy array as a georeferenced raster file (GeoTIFF by default).

    This function writes a 2D (single-band) or 3D (multi-band) NumPy array
    to a raster file using rasterio. Georeferencing can be applied by
    providing a CRS and an affine transform.

    Args:
        array (np.ndarray):
            Input array to save.
            - Shape (rows, cols): single-band raster.
            - Shape (bands, rows, cols): multi-band raster.
        output (str):
            Path to the output raster file.
        dtype (Optional[np.dtype], optional):
            Data type of the output raster. If None, inferred from the array.
            Defaults to None.
        compress (str, optional):
            Compression method for the raster (e.g., "lzw", "deflate").
            Defaults to "lzw".
        transpose (bool, optional):
            If True, assumes input shape is (bands, rows, cols) and transposes
            to (rows, cols, bands) before writing. Defaults to True.
        crs (Optional[str], optional):
            Coordinate reference system (e.g., "EPSG:4326").
            If None, no CRS is assigned. Defaults to None.
        transform (Optional[tuple], optional):
            Affine transform defining georeferencing.
            If None, raster is written without spatial reference.
            Defaults to None.
        driver (str, optional):
            GDAL driver to use for writing. Defaults to "GTiff".
        **kwargs:
            Additional keyword arguments passed to `rasterio.open()`.

    Returns:
        str: Path to the saved file.
    """
    # --- ensure correct shape ---
    if array.ndim == 3 and transpose:
        array = np.transpose(array, (1, 2, 0))

    # --- ensure output directory exists ---
    os.makedirs(os.path.dirname(os.path.abspath(output)), exist_ok=True)

    # --- get driver from extension ---
    ext = os.path.splitext(output)[-1].lower()
    driver_map = {"": "COG", ".tif": "GTiff", ".tiff": "GTiff", ".dat": "ENVI"}
    driver = driver_map.get(ext, "COG")
    if ext == "":
        output += ".tif"

    # --- infer dtype if not given ---
    if dtype is None:
        min_val, max_val = np.nanmin(array), np.nanmax(array)
        if min_val >= 0 and max_val <= 1:
            dtype = np.float32
        elif min_val >= 0 and max_val <= 255:
            dtype = np.uint8
        elif min_val >= -128 and max_val <= 127:
            dtype = np.int8
        elif min_val >= 0 and max_val <= 65535:
            dtype = np.uint16
        elif min_val >= -32768 and max_val <= 32767:
            dtype = np.int16
        else:
            dtype = np.float64
    array = array.astype(dtype)

    # --- set metadata ---
    count = 1 if array.ndim == 2 else array.shape[2]
    metadata = dict(
        driver=driver,
        height=array.shape[0],
        width=array.shape[1],
        count=count,
        dtype=array.dtype,
        crs=crs,
        transform=transform,
    )
    if compress and driver in ["GTiff", "COG"]:
        metadata["compress"] = compress
    metadata.update(**kwargs)

    # --- write raster ---
    with rasterio.open(output, "w", **metadata) as dst:
        if array.ndim == 2:  # panchromatic
            dst.write(array, 1)
            dst.set_band_description(
                1, kwargs.get("band_description", "Panchromatic band")
            )
        else:  # hyperspectral
            for i in range(array.shape[2]):
                dst.write(array[:, :, i], i + 1)
                if "wavelengths" in kwargs:
                    wl = kwargs["wavelengths"][i]
                    dst.set_band_description(i + 1, f"Band {i+1} ({wl:.1f} nm)")

    return output  # it's a file path


def read_prismaL2BC(
    file_path: str,
    product_type: str = "PRS_L2B",  # "PRS_L2B" or "PRS_L2C"
    wavelengths: Optional[List[float]] = None,
    method: str = "nearest",
    panchromatic: bool = False,
) -> Tuple[np.ndarray, Union[np.ndarray, str], np.ndarray, np.ndarray]:
    """
    Reads PRISMA Level-2B or Level-2C .he5 data (VNIR+SWIR or Panchromatic),
    apply scaling from DN to reflectance and return the data together with
    geolocation arrays.

    Args:
        file_path (str):
            Path to the PRISMA L2B/L2C `.he5` file.
        product_type (str, optional):
            PRISMA product type, either "PRS_L2B" or "PRS_L2C". Defaults to "PRS_L2B".
        wavelengths (Optional[List[float]], optional):
            List of target wavelengths in nanometers to extract (for hyperspectral).
            If None, all available wavelengths are used. Defaults to None.
        method (str, optional):
            Wavelength selection method:
                - "nearest": select the closest available band for each requested wavelength.
                - "exact": select only exact matches; raises ValueError if none found.
            Defaults to "nearest".
        panchromatic (bool, optional):
            If True, read the panchromatic cube instead of the hyperspectral VNIR+SWIR cubes.
            Defaults to False.

    Returns:
        Tuple[np.ndarray, Union[np.ndarray, str], np.ndarray, np.ndarray]:
            A 4-element tuple containing:
            - cube_or_pancube (np.ndarray):
                * Hyperspectral: 3D array with shape (rows, cols, wavelengths) of reflectance
                  (dtype float32) after applying scaling attributes.
                * Panchromatic: 2D array with shape (rows, cols) of reflectance.
            - wavelengths (np.ndarray or str):
                * Hyperspectral: 1D array of wavelengths (in nm) corresponding to the band axis.
                * Panchromatic: the string "Panchromatic".
            - lat (np.ndarray): 2D latitude array with shape (rows, cols).
            - lon (np.ndarray): 2D longitude array with shape (rows, cols).
    """
    if not check_valid_file(file_path, type=product_type):
        raise ValueError(
            f"{file_path} is not a valid {product_type} file or does not exist."
        )

    try:
        with h5py.File(file_path, "r") as f:
            if panchromatic:
                # --- PANCHROMATIC ---
                cube_path = f"HDFEOS/SWATHS/{product_type}_PCO/Data Fields/Cube"
                pancube_data = f[cube_path][()]

                fill_value = 0
                max_data_value = 65535
                l2_scale_pan_min = f.attrs.get("L2ScalePanMin", 0.0)
                l2_scale_pan_max = f.attrs.get("L2ScalePanMax", 1.0)

                pancube_data = l2_scale_pan_min + (
                    pancube_data.astype(np.float32) / max_data_value
                ) * (l2_scale_pan_max - l2_scale_pan_min)
                pancube_data[pancube_data == fill_value] = np.nan

                # --- read lat/lon ---
                lat = f[
                    f"HDFEOS/SWATHS/{product_type}_PCO/Geolocation Fields/Latitude"
                ][()]
                lon = f[
                    f"HDFEOS/SWATHS/{product_type}_PCO/Geolocation Fields/Longitude"
                ][()]

                wavelengths = "Panchromatic"
                return pancube_data, wavelengths, lat, lon

            else:
                # --- HYPERSPECTRAL VNIR+SWIR ---
                cube_path = f"HDFEOS/SWATHS/{product_type}_HCO/Data Fields"
                vnir_cube = f[f"{cube_path}/VNIR_Cube"][()]
                swir_cube = f[f"{cube_path}/SWIR_Cube"][()]
                vnir_wavelengths = f.attrs["List_Cw_Vnir"][()]
                swir_wavelengths = f.attrs["List_Cw_Swir"][()]

                max_data_value = 65535
                fill_value = 0

                l2_scale_vnir_min = f.attrs["L2ScaleVnirMin"][()]
                l2_scale_vnir_max = f.attrs["L2ScaleVnirMax"][()]
                vnir_cube = l2_scale_vnir_min + (
                    vnir_cube.astype(np.float32) / max_data_value
                ) * (l2_scale_vnir_max - l2_scale_vnir_min)

                l2_scale_swir_min = f.attrs["L2ScaleSwirMin"][()]
                l2_scale_swir_max = f.attrs["L2ScaleSwirMax"][()]
                swir_cube = l2_scale_swir_min + (
                    swir_cube.astype(np.float32) / max_data_value
                ) * (l2_scale_swir_max - l2_scale_swir_min)

                vnir_cube[vnir_cube == fill_value] = np.nan
                swir_cube[swir_cube == fill_value] = np.nan

                # --- combine cubes ---
                full_cube = np.concatenate((vnir_cube, swir_cube), axis=1)
                full_wavelengths = np.concatenate((vnir_wavelengths, swir_wavelengths))

                # --- filter and sort wavelengths ---
                valid_idx = full_wavelengths > 0
                full_wavelengths = full_wavelengths[valid_idx]
                full_cube = full_cube[:, valid_idx, :]
                sort_idx = np.argsort(full_wavelengths)
                full_wavelengths = full_wavelengths[sort_idx]
                full_cube = full_cube[:, sort_idx, :]

                # --- select requested wavelengths ---
                if wavelengths is not None:
                    requested = np.array(wavelengths)
                    if method == "exact":
                        idx = np.where(np.isin(full_wavelengths, requested))[0]
                        if len(idx) == 0:
                            raise ValueError(
                                "No requested wavelengths found (exact match)."
                            )
                    else:
                        idx = np.array(
                            [np.abs(full_wavelengths - w).argmin() for w in requested]
                        )
                    full_wavelengths = full_wavelengths[idx]
                    full_cube = full_cube[:, idx, :]

                full_cube = full_cube.transpose(0, 2, 1)
                # --- read lat/lon ---
                lat = f[
                    f"HDFEOS/SWATHS/{product_type}_HCO/Geolocation Fields/Latitude"
                ][()]
                lon = f[
                    f"HDFEOS/SWATHS/{product_type}_HCO/Geolocation Fields/Longitude"
                ][()]

                return full_cube, full_wavelengths, lat, lon

    except Exception as e:
        raise RuntimeError(f"Error reading the file {file_path}: {e}")


def write_prismaL2BC(
    filepath: str,
    output: str,
    product_type: str = "PRS_L2B",  # "PRS_L2B" or "PRS_L2C"
    panchromatic: bool = False,
    wavelengths: Optional[np.ndarray] = None,
    method: str = "nearest",
    **kwargs: Any,
) -> Optional[str]:
    """
    Convert a PRISMA Level-2B/2C dataset (.he5) into a georeferenced raster image.

    This function loads PRISMA hyperspectral (VNIR+SWIR) or panchromatic data,
    applies scaling and georeferencing using latitude/longitude fields, and writes
    the result to a raster file (e.g., GeoTIFF).

    Args:
        filepath (str):
            Path to the input PRISMA L2B/L2C `.he5` file.
        output (str):
            Path where the output raster will be saved.
        product_type (str, optional):
            PRISMA product type, either `"PRS_L2B"` or `"PRS_L2C"`.
            Defaults to `"PRS_L2B"`.
        panchromatic (bool, optional):
            If True, treat the dataset as a single-band panchromatic cube.
            If False, read the hyperspectral VNIR+SWIR cubes.
            Defaults to False.
        wavelengths (Optional[np.ndarray], optional):
            List/array of wavelengths (in nm) to select (hyperspectral only).
            If None, all available wavelengths are included.
            Defaults to None.
        method (str, optional):
            Method for wavelength selection:
                - `"nearest"`: select the closest available band for each requested wavelength.
                - `"exact"`: select only exact matches.
            Defaults to `"nearest"`.
        **kwargs (Any):
            Additional keyword arguments passed to:
                - `array_to_image()` for raster writing
                - `rasterio.open()` for file creation


    Returns:
        str: Output file path, or None if all values are NaN.
    """
    from affine import Affine

    # load dataset if it's a path to .he5
    if isinstance(filepath, str):
        if wavelengths is not None:
            full_cube, full_wavelengths, lat, lon = read_prismaL2BC(
                filepath,
                product_type=product_type,
                wavelengths=wavelengths,
                method=method,
                panchromatic=panchromatic,
            )
        else:
            full_cube, full_wavelengths, lat, lon = read_prismaL2BC(
                filepath,
                product_type=product_type,
                method=method,
                panchromatic=panchromatic,
            )

    # get np.array
    array = full_cube
    if not np.any(np.isfinite(array)):
        print("Warning: All reflectance values are NaN. Output image will be blank.")
        return None

    # get band names (wavelength) and, eventually, select specific bands
    if panchromatic:  # panchromatic
        kwargs["band_description"] = full_wavelengths
    else:  # cube
        kwargs["wavelengths"] = full_wavelengths

    # define transform
    res_y = 30 / 111320
    res_x = 30 / (111320 * np.cos(np.deg2rad(lat.mean())))

    x_min, x_max = lon.min() - res_x / 2, lon.max() + res_x / 2
    y_min, y_max = lat.min() - res_y / 2, lat.max() + res_y / 2
    extent = (x_min, y_min, x_max, y_max)

    ext_width = extent[2] - extent[0]
    ext_height = extent[3] - extent[1]
    height, width = lat.shape

    xResolution = ext_width / width
    yResolution = ext_height / height

    if np.allclose(lat[1:, 0] - lat[:-1, 0], lat[1, 0] - lat[0, 0]) and np.allclose(
        lon[0, 1:] - lon[0, :-1], lon[0, 1] - lon[0, 0]
    ):

        transform = Affine(xResolution, 0, extent[0], 0, -yResolution, extent[3])
    else:
        x0, y0 = lon[0, 0], lat[0, 0]
        x1, y1 = lon[0, 1], lat[0, 1]
        x2, y2 = lon[1, 0], lat[1, 0]

        dx_col = x1 - x0
        dy_col = y1 - y0
        dx_row = x2 - x0
        dy_row = y2 - y0

        transform = Affine(dx_col, dx_row, x0, dy_col, dy_row, y0)

    # write output
    return array_to_image(
        array,
        output=output,
        transpose=False,
        crs="EPSG:4326",
        transform=transform,
        **kwargs,
    )



"""The utils module provides functions used by the other modules."""

# TODO: get_transform assumes PRISMA L2D; does not work for L2B/L2C at the moment.

import os
import pyproj
from typing import List, Tuple
from affine import Affine


def check_valid_file(file_path: str, type: str = "PRS_L2D") -> bool:
    """
    Checks if the given file path points to a valid file.

    Args:
        file_path (str): Path to the file.
        type (str, optional): Expected file type ('PRS_L2B', 'PRS_L2C', 'PRS_L2D'). Defaults to 'PRS_L2D'.

    Returns:
        bool: True if file_path points to the correct file, False otherwise.
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"{file_path} does not exist.")

    valid_types = {"PRS_L2B", "PRS_L2C", "PRS_L2D"}
    if type not in valid_types:
        raise ValueError(
            f"Unsupported file type: {type}. Supported types are {valid_types}."
        )

    basename = os.path.basename(file_path)
    return basename.startswith(type) and basename.endswith(".he5")


def convert_coords(
    coords: List[Tuple[float, float]], from_epsg: str, to_epsg: str
) -> List[Tuple[float, float]]:
    """
    Convert a list of coordinates from one EPSG to another. Coordinates are
    expected in the format (lat, lon).

    Args:
        coords (List[Tuple[float, float]]):
            List of tuples containing coordinates in the format (latitude, longitude).
        from_epsg (str): Source EPSG code (e.g. "epsg:4326").
        to_epsg (str): Target EPSG code (e.g. "epsg:32632").

    Returns:
        List of tuple containing converted coordinates (x, y)
    """
    transformer = pyproj.Transformer.from_crs(from_epsg, to_epsg, always_xy=True)
    return [transformer.transform(lon, lat) for lat, lon in coords]


def get_transform(ul_easting: float, ul_northing: float, res: int = 30) -> Affine:
    """
    Returns an affine transformation for a given upper-left corner and resolution.

    Args:
        ul_easting (float): Easting coordinate of the upper-left corner.
        ul_northing (float): Northing coordinate of the upper-left corner.
        res (int, optional): Pixel resolution. Defaults to 30.

    Returns:
        Affine: Affine transformation object representing the spatial transform.
    """
    return Affine.translation(ul_easting, ul_northing) * Affine.scale(res, -res)



"""The spatial module provides functions to perform spatial analysis on PRISMA data products"""

import numpy as np
import xarray as xr
import rioxarray
from typing import Union, Tuple, Dict


def band_math(
    dataset: xr.Dataset,
    expressions: dict,
    wavelength_dict: dict,
    constants: dict = None,
    inplace: bool = False,
) -> xr.Dataset:
    """
    Compute one or more spectral indices from a hyperspectral dataset using
    user-defined expressions.

    Args:
        dataset (xarray.Dataset): Dataset containing the spectral variable
            (e.g., 'reflectance') with dimension 'wavelength'.
        expressions (dict): Dictionary mapping variable names to expressions,
            e.g. {'NDVI': '(nir - red) / (nir + red)', 'GNDVI': '(nir - green)/(nir + green)'}.
        wavelength_dict (dict): Dictionary mapping band names to wavelengths,
            e.g. {'red': 660, 'green': 560, 'nir': 840}.
        constants (dict, optional): Dictionary of constants to include in
            all expressions, e.g. {'c': 0.01}. Defaults to None.
        inplace (bool, optional):
            - If True, add the computed indices to the input dataset.
            - If False, return a new dataset containing only the indices.
            Defaults to False.

    Returns:
        xarray.Dataset: Dataset containing the computed index/indices.
            If inplace=True, also keeps the original variables.

    Example:
        >>> wavelength_dict = {'red': 660, 'green': 560, 'nir': 840}
        >>> expressions = {
        ...     'NDVI': '(nir - red) / (nir + red)',
        ...     'GNDVI': '(nir - green) / (nir + green)'
        ... }
        >>> result = band_math(ds, expressions, wavelength_dict, inplace=True)
    """
    da_dict = {
        band: dataset["reflectance"].sel(wavelength=w, method="nearest")
        for band, w in wavelength_dict.items()
    }

    if constants:
        da_dict.update(constants)

    results = {}
    for varname, expr in expressions.items():
        results[varname] = eval(expr, {"__builtins__": None}, da_dict)

    if inplace:
        for varname, da in results.items():
            dataset[varname] = da
        return dataset
    else:
        return xr.Dataset(results)


def pca_image(
    input_data: Union[xr.Dataset, str], out_tif: str, n_components: int = 3
) -> Tuple[xr.Dataset, Dict[str, str]]:
    """
    Perform Principal Component Analysis (PCA) on a hyperspectral or multispectral image
    and save the resulting principal components as a GeoTIFF and an xarray.Dataset.

    This function supports both:
      - `xarray.Dataset` input containing a 'reflectance' variable with dimensions
        (y, x, wavelength)
      - Filepath to a GeoTIFF with shape (bands, rows, cols)

    The function reshapes the image to (pixels, bands), applies PCA, and reshapes the
    principal components back to image dimensions. It also preserves geospatial metadata
    (CRS and transform) when saving the output GeoTIFF.

    Args:
        input_data (xarray.Dataset or str):
            - xarray.Dataset: Must contain 'reflectance' variable with shape (y, x, wavelength)
            - str: Filepath to a GeoTIFF
        out_tif (str): Output path to save the PCA GeoTIFF.
        n_components (int, optional): Number of principal components to compute. Default is 3.

    Returns:
        ds_pca (xarray.Dataset): Dataset containing the principal components with dimensions
            ("pc", "y", "x") and coordinates:
              - pc: principal component index (1, 2, ..., n_components)
              - y: spatial y-coordinate
              - x: spatial x-coordinate
        expl_var (dict): Explained variance ratio for each principal component, formatted as
            percentages with two decimals, e.g. {"PC1": "78.32", "PC2": "12.45", ...}

    Example:
        >>> # Using xarray Dataset
        >>> ds_pca, expl_var = pca_image(ds, "output/pca_image.tif", n_components=3)
        >>> print(expl_var)
        {'PC1': '78.32', 'PC2': '12.45', 'PC3': '5.67'}

        >>> # Using GeoTIFF
        >>> ds_pca, expl_var = pca_image("input_image.tif", "output/pca_image.tif", n_components=5)

    Notes:
        - For xarray input, the function expects the CRS and transform to be set with `rioxarray`.
        - The PCA output GeoTIFF will have `n_components` bands corresponding to the principal
          components in order.
        - The explained variance ratio provides insight into how much of the data variability
          is captured by each component.
    """
    from sklearn.decomposition import PCA
    import rasterio

    # read data
    if isinstance(input_data, xr.Dataset):
        ds = input_data
        array = ds["reflectance"].values  # rows, cols, bands
        array = np.transpose(array, (2, 0, 1))  # bands, rows, cols
        bands, rows, cols = array.shape
        x_coords = ds["x"].values
        y_coords = ds["y"].values
        crs = ds.rio.crs
        geotransform = ds.rio.transform()

    elif isinstance(input_data, str):
        with rasterio.open(input_data) as src:
            array = src.read()  # shape: bands, rows, cols
            bands, rows, cols = array.shape
            crs = src.crs
            geotransform = src.transform
            x_coords = np.array([(geotransform * (col, 0))[0] for col in range(cols)])
            y_coords = np.array([(geotransform * (0, row))[1] for row in range(rows)])

    else:
        raise ValueError("input_data must be xr.Dataset or a GeoTIFF filepath")

    # PCA
    image_reshaped = array.reshape(bands, rows * cols).T  # n_samples, n_bands
    model = PCA(n_components=n_components)
    principal_components = model.fit_transform(image_reshaped)
    pca_image = principal_components.T.reshape(n_components, rows, cols)

    expl_var = {
        f"PC{i+1}": f"{v*100:.2f}"
        for i, v in enumerate(model.explained_variance_ratio_)
    }

    # write output
    ds_pca = xr.Dataset(
        data_vars=dict(
            pca=(["pc", "y", "x"], pca_image),
        ),
        coords=dict(
            pc=(["pc"], np.arange(1, n_components + 1)),
            y=(["y"], y_coords),
            x=(["x"], x_coords),
        ),
    )
    ds_pca.rio.write_crs(crs, inplace=True)
    ds_pca.rio.write_transform(geotransform, inplace=True)

    metadata = dict(
        driver="GTiff",
        height=pca_image.shape[1],
        width=pca_image.shape[2],
        count=pca_image.shape[0],
        dtype=pca_image.dtype,
        crs=crs,
        transform=geotransform,
    )

    with rasterio.open(out_tif, "w", **metadata) as dst:
        dst.write(pca_image)
        print(f"pca image saved at: {out_tif}")

    return ds_pca, expl_var



"""Main module."""

import numpy as np
import tempfile

import ipyleaflet
import leafmap

from .prisma import read_prismaL2D, write_prismaL2D
from .gui import main_toolbar, spectralWidget


class Map(leafmap.Map):
    """
    A class that extends leafmap.Map to provide additional functionality for
        prismatools.

    Attributes:
        Any attributes inherited from leafmap.Map.

    Methods:
        Any methods inherited from leafmap.Map.
    """

    def __init__(self, center=[0, 0], zoom=2, **kwargs):
        """
        Initializes a new instance of the Map class.

        Args:
            **kwargs: Arbitrary keyword arguments that are passed to the parent
                class's constructor.
        """
        # set off toolbars not needed from Leafmap
        kwargs["layers_control"] = False
        kwargs["fullscreen_control"] = True
        kwargs["search_control"] = False
        kwargs["draw_control"] = False
        kwargs["measure_control"] = False
        kwargs["scale_control"] = True
        kwargs["toolbar_control"] = False

        super().__init__(**kwargs)
        # add my simplified toolbar
        main_toolbar(self)

        self._spectral_data = {}
        self._plot_options = None

    def add(self, obj, position="topright", xlim=None, ylim=None, **kwargs):
        """Add a layer to the map.

        Args:
            obj (str or object): The name of the layer or a layer object.
            position (str, optional): The position of the layer widget. Can be
                'topright', 'topleft', 'bottomright', or 'bottomleft'. Defaults
                to 'topright'.
            xlim (tuple, optional): The x-axis limits of the plot. Defaults to None.
            ylim (tuple, optional): The y-axis limits of the plot. Defaults to None.
            **kwargs: Arbitrary keyword arguments that are passed to the parent
                class's add_layer method.
        """

        if isinstance(obj, str):
            if obj == "spectral":

                spectralWidget(self, position=position, xlim=xlim, ylim=ylim, **kwargs)
                self.set_plot_options(add_marker_cluster=True)
            else:
                super().add(obj, **kwargs)

        else:
            super().add(obj, **kwargs)

    def add_prisma(
        self,
        source,
        wavelengths=None,
        indexes=None,
        colormap=None,
        vmin=0,
        vmax=0.5,
        nodata=np.nan,
        attribution=None,
        layer_name="PRISMA",
        zoom_to_layer=True,
        visible=True,
        array_args=None,
        method="nearest",
        **kwargs,
    ):
        """Add a PRISMA dataset to the map.

        This function reads a PRISMA hyperspectral dataset, optionally selects
        specific wavelengths, converts the data to an image, and adds it as a
        raster layer to the map. The dataset can be provided as a file path or
        as an xarray Dataset.

        Args:
            source (str or xarray.Dataset): The path to the PRISMA file or an
                in-memory xarray Dataset containing PRISMA data.
            wavelengths (list or np.ndarray, optional): Specific wavelengths to
                select from the dataset. If None, all wavelengths are used.
                Defaults to None.
            indexes (int or list, optional): The band(s) to display. Band
                indexing starts at 1. Defaults to None.
            colormap (str, optional): The name of the colormap from `matplotlib`
                to use when plotting a single band. See:
                https://matplotlib.org/stable/gallery/color/colormap_reference.html.
                Default is greyscale.
            vmin (float, optional): The minimum value for color mapping when
                plotting a single band. Defaults to 0.
            vmax (float, optional): The maximum value for color mapping when
                plotting a single band. Defaults to 0.5.
            nodata (float, optional): Value in the raster to interpret as
                no-data. Defaults to np.nan.
            attribution (str, optional): Attribution for the source raster.
                Defaults to None.
            layer_name (str, optional): The name to assign to the map layer.
                Defaults to "PRISMA".
            zoom_to_layer (bool, optional): Whether to zoom the map to the
                extent of the layer after adding it. Defaults to True.
            visible (bool, optional): Whether the layer should be visible when
                first added. Defaults to True.
            array_args (dict, optional): Additional keyword arguments to pass to
                `array_to_memory_file` when reading the raster. Defaults to {}.
            method (str, optional): Method to use for wavelength interpolation
                when selecting bands. Options may include "nearest", "linear",
                etc. Defaults to "nearest".
            **kwargs: Additional keyword arguments passed to `add_raster`.
        """
        if array_args is None:
            array_args = {}

        if isinstance(source, str):
            xds = read_prismaL2D(source)
        else:
            xds = source

        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp:
            temp_path = tmp.name

        # create a temporary .tif with only selected wavelengths because leafmap (built on ipyleaflet) can add only images with 3 bands or 1 band
        write_prismaL2D(xds, wavelengths=wavelengths, method=method, output=temp_path)

        self.add_raster(  # this function will add the tiff to the map
            temp_path,
            indexes=indexes,
            colormap=colormap,
            vmin=vmin,
            vmax=vmax,
            nodata=nodata,
            attribution=attribution,
            layer_name=layer_name,
            zoom_to_layer=zoom_to_layer,
            visible=visible,
            array_args=array_args,
            **kwargs,
        )

        self.cog_layer_dict[layer_name][
            "xds"
        ] = xds  # avoid losing spectral information
        self.cog_layer_dict[layer_name]["hyper"] = "PRISMA"

    def add_raster(
        self,
        source,
        indexes=None,
        colormap=None,
        vmin=None,
        vmax=None,
        nodata=None,
        attribution=None,
        layer_name="Raster",
        layer_index=None,
        zoom_to_layer=True,
        visible=True,
        opacity=1.0,
        array_args=None,
        client_args={"cors_all": False},
        open_args=None,
        **kwargs,
    ):
        """Add a local raster dataset to the map.
            If you are using this function in JupyterHub on a remote server
                (e.g., Binder, Microsoft Planetary Computer) and
            if the raster does not render properly, try installing
                jupyter-server-proxy using `pip install jupyter-server-proxy`,
            then running the following code before calling this function. For
                more info, see https://bit.ly/3JbmF93.

            import os
            os.environ['LOCALTILESERVER_CLIENT_PREFIX'] = 'proxy/{port}'

        Args:
            source (str): The path to the GeoTIFF file or the URL of the Cloud
                Optimized GeoTIFF.
            indexes (int, optional): The band(s) to use. Band indexing starts
                at 1. Defaults to None.
            colormap (str, optional): The name of the colormap from `matplotlib`
                to use when plotting a single band. See
                https://matplotlib.org/stable/gallery/color/colormap_reference.html.
                Default is greyscale.
            vmin (float, optional): The minimum value to use when colormapping
                the palette when plotting a single band. Defaults to None.
            vmax (float, optional): The maximum value to use when colormapping
                the palette when plotting a single band. Defaults to None.
            nodata (float, optional): The value from the band to use to interpret
                as not valid data. Defaults to None.
            attribution (str, optional): Attribution for the source raster. This
                defaults to a message about it being a local file.. Defaults to None.
            layer_name (str, optional): The layer name to use. Defaults to 'Raster'.
            layer_index (int, optional): The index of the layer. Defaults to None.
            zoom_to_layer (bool, optional): Whether to zoom to the extent of the
                layer. Defaults to True.
            visible (bool, optional): Whether the layer is visible. Defaults to
                True.
            opacity (float, optional): The opacity of the layer. Defaults to 1.0.
            array_args (dict, optional): Additional arguments to pass to
                `array_to_memory_file` when reading the raster. Defaults to {}.
            client_args (dict, optional): Additional arguments to pass to
                localtileserver.TileClient. Defaults to { "cors_all": False }.
            open_args (dict, optional): Additional arguments to pass to
                rioxarray.open_rasterio.

        """
        import rioxarray as rxr

        if array_args is None:
            array_args = {}
        if open_args is None:
            open_args = {}

        if nodata is None:
            nodata = np.nan

        super().add_raster(
            source,
            indexes=indexes,  # if source has more bands, you need to specify 3 indices
            colormap=colormap,
            vmin=vmin,
            vmax=vmax,
            nodata=nodata,
            attribution=attribution,
            layer_name=layer_name,
            layer_index=layer_index,
            zoom_to_layer=zoom_to_layer,
            visible=visible,
            opacity=opacity,
            array_args=array_args,
            client_args=client_args,
            **kwargs,
        )

        if isinstance(source, str):
            da = rxr.open_rasterio(source, **open_args)
            dims = da.dims
            da = da.transpose(dims[1], dims[2], dims[0])

            xds = da.to_dataset(name="data")
            self.cog_layer_dict[layer_name]["xds"] = xds

    def spectral_to_df(self, **kwargs):
        """Converts the spectral data to a pandas DataFrame.

        Returns:
            pd.DataFrame: The spectral data as a pandas DataFrame.
        """
        import pandas as pd

        df = pd.DataFrame(self._spectral_data, **kwargs)
        return df

    def spectral_to_csv(self, filename, index=True, **kwargs):
        """Saves the spectral data to a CSV file.

        Args:
            filename (str): The output CSV file.
            index (bool, optional): Whether to write the index. Defaults to True.
        """
        df = self.spectral_to_df()
        df = df.rename_axis("band")
        df.to_csv(filename, index=index, **kwargs)

    def set_plot_options(
        self,
        add_marker_cluster=False,
        plot_type=None,
        overlay=False,
        position="bottomright",
        min_width=None,
        max_width=None,
        min_height=None,
        max_height=None,
        **kwargs,
    ):
        """Sets plotting options.

        Args:
            add_marker_cluster (bool, optional): Whether to add a marker cluster.
                Defaults to False.
            sample_scale (float, optional):  A nominal scale in meters of the
                projection to sample in . Defaults to None.
            plot_type (str, optional): The plot type can be one of "None", "bar",
                "scatter" or "hist". Defaults to None.
            overlay (bool, optional): Whether to overlay plotted lines on the
                figure. Defaults to False.
            position (str, optional): Position of the control, can be
                ‘bottomleft’, ‘bottomright’, ‘topleft’, or ‘topright’. Defaults
                to 'bottomright'.
            min_width (int, optional): Min width of the widget (in pixels), if
                None it will respect the content size. Defaults to None.
            max_width (int, optional): Max width of the widget (in pixels), if
                None it will respect the content size. Defaults to None.
            min_height (int, optional): Min height of the widget (in pixels), if
                None it will respect the content size. Defaults to None.
            max_height (int, optional): Max height of the widget (in pixels), if
                None it will respect the content size. Defaults to None.

        """
        plot_options_dict = {}
        plot_options_dict["add_marker_cluster"] = add_marker_cluster
        plot_options_dict["plot_type"] = plot_type
        plot_options_dict["overlay"] = overlay
        plot_options_dict["position"] = position
        plot_options_dict["min_width"] = min_width
        plot_options_dict["max_width"] = max_width
        plot_options_dict["min_height"] = min_height
        plot_options_dict["max_height"] = max_height

        for key in kwargs:
            plot_options_dict[key] = kwargs[key]

        self._plot_options = plot_options_dict

        if not hasattr(self, "_plot_marker_cluster"):
            self._plot_marker_cluster = ipyleaflet.MarkerCluster(name="Marker Cluster")

        if add_marker_cluster and (self._plot_marker_cluster not in self.layers):
            self.add(self._plot_marker_cluster)
